require "blockify/version"

module Blockify
  def blockify_elements(&block)  # blockify
    if self.respond_to? :each_pair
      hash = {}
      self.each_pair do |key,val|
        if val.respond_to? :each
          hash[key] = val.blockify_elements &block
        else
          hash[key] = block.call(val)
        end   
      end
      return hash
    elsif self.respond_to? :each
      ary = []
      self.each do |val|
        if val.respond_to? :each
          ary.push val.blockify_elements &block
        else
          ary.push block.call(val)
        end   
      end
      return ary
    end
    return self # should never get here
  end
  
  def find_element_path(path=[],done=[false], &block)
    if self.respond_to? :each_pair
      self.each_pair do |key,val|
        if val.respond_to? :each
          path.push key
          val.find_element_path(path, done, &block)
          return path.self_or_empty_as_nil if done.first
          path.pop
        else
          if block.call(val)
            path.push key
            done[0]=true
            return path.self_or_empty_as_nil
          end
        end
      end
    elsif self.respond_to? :each
      idx = 0
      self.each do |val|
        if val.respond_to? :each
          path.push idx
          val.find_element_path(path, done, &block)
          return path.self_or_empty_as_nil if done.first
          path.pop
        else
          if block.call(val)
            path.push idx
            done[0]=true
            return path.self_or_empty_as_nil
          end
        end
        idx += 1
      end
    end # if-else
    return path.self_or_empty_as_nil
  end

  def find_element_paths(path=[],paths=[], &block)
    if self.respond_to? :each_pair
      self.each_pair do |key,val|
        if val.respond_to? :each
          path.push key
          val.find_element_paths(path, paths, &block)
          path.pop
        else
          if block.call(val)
            path.push key
            paths.push path.dup
            path.pop
          end
        end
      end
    elsif self.respond_to? :each
      idx = 0
      self.each do |val|
        if val.respond_to? :each
          path.push idx
          val.find_element_paths(path, paths, &block)
          path.pop
        else
          if block.call(val)
            path.push idx
            paths.push path.dup
            path.pop
          end
        end
        idx += 1
      end
    end # if-else
    return paths
  end
  
  def path_get(path)
    item = self
    path.each do |idx|
      item = item[idx]
    end
    return item
  end
  
  def paths_get(paths)
    rtn = []
    paths.each do |path|
      rtn.push path_get(path)
    end
    return rtn
  end
  
  def path_put(val,path)
    idx = path.pop
    con = path_get(path)
    rtn = con[idx]
    con[idx]=val
    path.push idx # put back
    return rtn
  end
  
  # read elements and yield everything
  def scan_elements(&block)
    if self.respond_to? :each_pair
      self.each_pair do |key,val|
        if val.respond_to? :each
          val.scan_elements(&block)
        else
          block.call(key,self)
        end
      end
    elsif self.respond_to? :each
      ii = 0
      self.each do |val|
        if val.respond_to? :each
          val.scan_elements(&block)
        else
          block.call(ii,self)
        end
        ii += 1
      end    
    end 
  end
  
  # add path later to see where things break down
  def circular? (path = [], stack = {}, rtn = [false])
    my_id = self.object_id
    stack.incrementify(my_id)
    if self.respond_to? :each_pair
      self.each_pair do |key,val|
        if val.respond_to? :each
          x_id = val.object_id
          if stack[x_id]
            rtn[0]=true
            return rtn.first
          else
            path.push key
            return rtn.first if val.circular?(path, stack, rtn)
            path.pop
          end
        end
      end
    elsif self.respond_to? :each    
      ii = 0
      self.each do |val|
        if val.respond_to? :each
          x_id = val.object_id
          if stack[x_id]
            rtn[0]=true
            path.push ii
            return rtn.first
          else
            path.push ii
              return rtn.first if val.circular?(path, stack, rtn)
            path.pop  
          end
        end
        ii += 1
      end
    end
    stack.decrementify(my_id)
    rtn.first
  end
  
  def unloopify
    path = []
    while circular? path do
      obj = path_get path
      nobj = CircularObjectContainer.new obj
      path_put(nobj, path)
      path = []
    end
  end
  
  def flattenify
    rtn = []
    scan_elements { |key, inst| rtn.push inst[key] }
    return rtn
  end
  
  def extractify(cls)
    col = cls.new
    scan_elements do |key, inst|
      col[key] = inst[key] if (inst.kind_of? cls)
    end
    return col    
  end
  
  def extractifirstify(cls)
    col = cls.new
    scan_elements do |key, inst|
      if (inst.kind_of? cls)
        if col[key].nil?
          col[key] = inst[key]
        end
      end
    end
    return col    
  end

  def extractistacktify(cls)
    col = cls.new
    scan_elements do |key, inst|
      if (inst.kind_of? cls)
        col[key] ||= []
        col[key].push inst[key]
      end
    end
    return col    
  end
  
  def includify?(search_string)
    path = find_element_path { |elm| elm.to_s.include? search_string }
    !path.nil?
  end
  
  def blockify_elements!(&block)
    replace blockify_elements &block
  end
  def stringify_elements
    blockify_elements {|t| t.to_s}
  end
  def stringify_elements!
    replace stringify_elements
  end
  def inspectify_elements
    blockify_elements {|t| t.inspect}
  end
  def inspectify_elements!
    replace inspectify_elements
  end
  
  def deep_duplify
    blockify_elements do |t|
      t.dup rescue t
    end
  end
  
  def self_or_empty_as_nil
    return nil if empty?
    self
  end

end # module

Array.include Blockify
Hash.include Blockify

class Hash
  def incrementify(kv) #values assumed to be integers
    if self[kv].nil?
      self[kv] = 0
    else
      self[kv] += 1
    end
  end
  def decrementify(kv) #values assumed to be integers
    if self[kv].nil?
      self[kv] = nil 
    elsif self[kv]==0
      self[kv] = nil      
    else
      self[kv] -= 1
    end
  end
end

class Object
  def taboo?
    false
  end
end

class CircularObjectContainer
  def initialize(obj)
    @id = obj.object_id
    if obj.respond_to? :each_pair
      @obj = obj.class.new  # create a blank slate
      obj.each_pair do |key,val|
        if val.respond_to? :each
          @obj[key] = val.class.new  # go no further
        else
          tval = val.dup  rescue val
          @obj[key] = tval
        end
      end
      self.instance_eval do # yes we want a singleton method here!
        def each_pair(&block)
          @obj.each_pair &block
        end
        def each(&block)
          @obj.each &block
        end
        def [](key)
          @obj[key]
        end
      end 
    elsif obj.respond_to? :each
      @obj = obj.class.new  # create a blank slate
      obj.each do |val|
        if val.respond_to? :each
          @obj.push val.class.new
        else
          tval = val.dup rescue val
          @obj.push tval
        end
      end
      self.instance_eval do  # yes we want a singleton method here!
        def each(&block)
          @obj.each &block
        end
        def [](idx)
          @obj[idx]
        end        
      end 
    else
      @obj = obj.dup rescue obj  
    end
    @obj.freeze
    self.freeze
    self
  end  
  def id  # get the original object_id
    @id
  end
  def type
    @obj.class
  end
  def dup  # get the innards and discard container
    @obj.dup rescue @obj
  end
  def taboo?
    true
  end
  def circular? (path = [], stack = {}, rtn = [false])
    return rtn.first
  end
  def blockify_elements(&block)
    @obj.dup.blockify_elements(&block) 
  end
  def find_element_path(path=[],done=[false], &block)
    @obj.dup.find_element_path(path,done, &block) 
  end
  def find_element_paths(path=[],paths=[], &block)
    @obj.dup.find_element_paths(path,paths, &block)
  end
  def scan_elements(&block)
    @obj.dup.scan_elements(&block)
  end
end


